import { StyleMapping, TemplateAnalysis, TemplateIntegrationOptions, TemplateTypes } from "@opticss/template-api";
import { Actions } from "./Actions";
import { CssFile } from "./CssFile";
import { OptiCSSOptions } from "./OpticssOptions";
export interface OptimizationResult {
    output: CssFile;
    styleMapping: StyleMapping;
    actions: Actions;
}
export interface TimingData {
    [timingName: string]: {
        start: number;
        end: number;
    };
}
export declare class Optimizer {
    sources: Array<CssFile>;
    analyses: Array<TemplateAnalysis<keyof TemplateTypes>>;
    options: OptiCSSOptions;
    templateOptions: TemplateIntegrationOptions;
    timings: TimingData;
    private initializers;
    private optimizations;
    /**
     * Creates a new OptiCSS Optimizer.
     *
     * @param sources a list of css files to be optimized.
     *   within a given css file, the cascade is respected as a conflict resolution
     *   signal. Classes from multiple files are assumed to have an arbitrary ordering
     *   and the cascade is not used to resolve conflicts between properties. Instead,
     *   those conflicts must be resolvable by having analysis information that proves
     *   they don't conflict or by having selectors that unambiguously resolve the conflict.
     */
    constructor(options: Partial<OptiCSSOptions>, templateOptions: Partial<TemplateIntegrationOptions>);
    /**
     * Add another source file to include in this optimization.
     * @param CSS File to add
     */
    addSource(file: CssFile): void;
    /**
     * Add another TemplateAnalysis to use in this optimization.
     * @param TemplateAnalysis to use.
     */
    addAnalysis(analysis: TemplateAnalysis<keyof TemplateTypes>): void;
    /**
     * Utility method to save timing data.
     * @param Timing measurement name.
     * @param Timing measurement start.
     * @param Timing measurement end.
     */
    private logTiming;
    /**
     * Parse all CSS files this optimizer is concerned with.
     * @param All sources registered with this Optimizer.
     * @returns All files we're working on, now represented as postcss trees.
     */
    private parseFiles;
    /**
     * Run all optimizations' initializers.
     * @param This optimization pass instance.
     * @param All files we're working on, parsed as postcss trees.
     */
    private initialize;
    /**
     * Run all of this optimizer's optimizations.
     * @param This optimization pass instance.
     * @param All files we're working on, parsed as postcss trees.
     * @returns All the files we're just transformed.
     */
    private optimizeFiles;
    /**
     * Concatenate all of this Optimizer's files into a single output.
     * @param All postcss ASTs we're working on.
     * @param The expected output's filename.
     * @returns The concatenated file.
     */
    private concatenateFiles;
    /**
     * Main runner method for the Optimizer. After all Sources and Analyses are registered,
     * calling this method executes all requested optimizations and returns an optimization result.
     * @param outputFilename - The output's filename. The file is not written
     *   but it is needed to ensure that source maps works correctly.
     * @returns The optimization result.
     */
    optimize(outputFilename: string): Promise<OptimizationResult>;
}
