import * as postcss from "postcss";
import * as postcssSelectorParser from "postcss-selector-parser";
export interface CombinatorAndSelector<SelectorType> {
    combinator: postcssSelectorParser.Combinator;
    selector: SelectorType;
}
export declare type CombinatorAndCompoundSelector = CombinatorAndSelector<CompoundSelector>;
export declare class CombinedSelector<T> {
    next?: CombinatorAndSelector<T>;
    setNext(combinator: postcssSelectorParser.Combinator, selector: T): void;
}
/**
 * Core structure in `CompoundSelector` linked lists. Keeps reference to the next
 * `CompoundSelector` in the chan, and the combinator that connects them.
 */
export declare class CompoundSelector extends CombinedSelector<CompoundSelector> {
    nodes: postcssSelectorParser.Node[];
    pseudoelement?: postcssSelectorParser.Pseudo;
    next?: CombinatorAndCompoundSelector;
    constructor();
    /**
     * Crawl the linked list and return last sibling.
     * @return The last `CompoundSelector` in list
     */
    readonly lastSibling: CompoundSelector;
    /**
     * Add a node to `CompoundSelector`.
     * @param Simple selector node to add.
     */
    addNode(node: postcssSelectorParser.Node): void;
    /**
     * Set pseudo element type on this `CompoundSelector`.
     * @param The `selectorParser.Pseudo` to assign.
     */
    setPseudoelement(pseudo: postcssSelectorParser.Pseudo): void;
    /**
     * Insert a new `CompoundSelector` before another in the linked list.
     * @param other The `CompoundSelector` to insert.
     * @param combinator The combinator connecting these two selectors.
     * @param reference The `CompoundSelector` to insert this selector before.
     */
    insertBefore(other: CompoundSelector, combinator: postcssSelectorParser.Combinator, reference: CompoundSelector): boolean;
    /**
     * Append a new `CompoundSelector` to the linked list.
     * @param combinator The combinator connecting the new selectors.
     * @param selector The `CompoundSelector` to append.
     * @return This `CompoundSelector`.
     */
    append(combinator: postcssSelectorParser.Combinator, selector: CompoundSelector): CompoundSelector;
    /**
     * Merge this `CompoundSelector` with another.
     * @param other The `CompoundSelector` to merge with this one.
     * @return This `CompoundSelector`.
     */
    mergeNodes(other: CompoundSelector): CompoundSelector;
    /**
     * Remove the last `CompoundSelector` in this linked list.
     * @return The removed `CompoundSelector`, or undefined.
     */
    removeLast(): CombinatorAndCompoundSelector | undefined;
    /**
     * Clone the `CompoundSelector` linked list starting at this node.
     * @return The cloned `CompoundSelector` linked list.
     */
    clone(): CompoundSelector;
    /**
     * Stringify this `CompoundSelector` list back into CSS.
     * @return The selector string.
     */
    toString(shallow?: boolean): string;
}
/**
 * `ParsedSelector` serves as a container object for a `CompoundSelector` linked
 * list and provides a number of convenience methods for interacting with it.
 */
export declare class ParsedSelector {
    /**
     * The raw string the selector was parsed from, if any.
     */
    source: string | undefined;
    private _key;
    selector: CompoundSelector;
    /**
     * @param Root `CompoundSelector` of linked list to track.
     */
    constructor(selector: CompoundSelector, source?: string);
    eachCompoundSelector<EarlyReturnType>(callback: (sel: CompoundSelector) => EarlyReturnType | undefined): EarlyReturnType | undefined;
    eachSelectorNode<EarlyReturnType>(callback: (node: postcssSelectorParser.Node) => EarlyReturnType | undefined): EarlyReturnType | undefined;
    /**
     * Checks if a given `CompoundSelector` is a context selector for this `ParsedSelector`.
     * @return True or false depending on if is context selector.
     */
    isContext(selector: CompoundSelector): boolean;
    /**
     * Returns the key selector (last compound selector) of this selector.
     * @return Key selector.
     */
    readonly key: CompoundSelector;
    /**
     * Returns the number of `CompoundSelector` present in this `ParsedSelector`
     * @return CompoundSelector count.
     */
    readonly length: number;
    /**
     * Returns a deep clone of this `ParsedSelector` and the linked list it tracks.
     * @return new `ParsedSelector` clone.
     */
    clone(): ParsedSelector;
    /**
     * Stringify this `CompoundSelector` list back into CSS.
     * @return The selector string.
     */
    toString(): string;
    toContext(...keyNodes: Array<postcssSelectorParser.Node>): ParsedSelector;
    /**
     * Same as toString() but having omitted the key selector.
     * Includes the final combinator prior to the key selector.
     */
    toContextString(): string;
}
/**
 * All valid selector-like inputs to the `parseSelector` helper methods.
 */
export declare type Selectorish = string | postcssSelectorParser.Root | postcssSelectorParser.Node[] | postcssSelectorParser.Node[][] | postcss.Rule;
/**
 * Converts a selector like object, return an array of `CompoundSelector` objects.
 *
 * @param selector  Selector like object: including `string`, `selectorParser.Root`, `selectorParser.Selector`, or `selectorParser.Node`
 * @return Array of `CompoundSelector` objects.
 */
export declare function parseCompoundSelectors(selector: Selectorish): CompoundSelector[];
/**
 * Given a selector like object, return an array of `ParsedSelector` objects.
 * Passed `selector` may be any valid selector as a string, or `postcss-selector-parser`
 * object.
 *
 * @param selector  Selector like object: including `string`, `selectorParser.Root`, `selectorParser.Selector`, or `selectorParser.Node`
 * @return Array of `ParsedSelector` objects.
 */
export declare function parseSelector(selector: Selectorish): ParsedSelector[];
